@@ -3,25 +3,21 @@
 #include "config.h"
 #include "device.h"
 #include "socket.h"
-#include "packets.h"
+#include "queueing.h"
 #include "timers.h"
 #include "hashtables.h"
 #include "peer.h"
 #include "uapi.h"
 
-static int clear_peer_endpoint_src(struct wireguard_peer *peer, void *data)
-{
-	socket_clear_peer_endpoint_src(peer);
-	return 0;
-}
-
 static int set_device_port(struct wireguard_device *wg, u16 port)
 {
+	struct wireguard_peer *peer, *temp;
 	socket_uninit(wg);
 	wg->incoming_port = port;
-	if (!(netdev_pub(wg)->flags & IFF_UP))
+	if (!(wg->dev->flags & IFF_UP))
 		return 0;
-	peer_for_each_unlocked(wg, clear_peer_endpoint_src, NULL);
+	peer_for_each (wg, peer, temp, false)
+		socket_clear_peer_endpoint_src(peer);
 	return socket_init(wg);
 }
 
@@ -60,11 +56,23 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 	peer = pubkey_hashtable_lookup(&wg->peer_hashtable, in_peer.public_key);
 	if (!peer) { /* Peer doesn't exist yet. Add a new one. */
 		if (in_peer.flags & WGPEER_REMOVE_ME)
-			return -ENODEV; /* Tried to remove a non existing peer. */
-		peer = peer_rcu_get(peer_create(wg, in_peer.public_key));
+			return -ENODEV; /* Tried to remove a non-existing peer. */
+		if (in_peer.flags & WGPEER_REMOVE_PRESHARED_KEY)
+			return -EINVAL; /* Tried to remove a psk for a non-existing peer. */
+
+		down_read(&wg->static_identity.lock);
+		if (wg->static_identity.has_identity && !memcmp(in_peer.public_key, wg->static_identity.static_public, NOISE_PUBLIC_KEY_LEN)) {
+			/* We silently ignore peers that have the same public key as the device. The reason we do it silently
+			 * is that we'd like for people to be able to reuse the same set of API calls across peers. */
+			up_read(&wg->static_identity.lock);
+			goto out;
+		}
+		up_read(&wg->static_identity.lock);
+
+		peer = peer_rcu_get(peer_create(wg, in_peer.public_key, in_peer.preshared_key));
 		if (!peer)
 			return -ENOMEM;
-		if (netdev_pub(wg)->flags & IFF_UP)
+		if (wg->dev->flags & IFF_UP)
 			timers_init_peer(peer);
 	}
 
@@ -74,6 +82,16 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 		goto out;
 	}
 
+	if (in_peer.flags & WGPEER_REMOVE_PRESHARED_KEY) {
+		down_write(&peer->handshake.lock);
+		memset(&peer->handshake.preshared_key, 0, NOISE_SYMMETRIC_KEY_LEN);
+		up_write(&peer->handshake.lock);
+	} else if (memcmp(zeros, in_peer.preshared_key, WG_KEY_LEN)) {
+		down_write(&peer->handshake.lock);
+		memcpy(&peer->handshake.preshared_key, in_peer.preshared_key, NOISE_SYMMETRIC_KEY_LEN);
+		up_write(&peer->handshake.lock);
+	}
+
 	if (in_peer.endpoint.addr.sa_family == AF_INET || in_peer.endpoint.addr.sa_family == AF_INET6) {
 		struct endpoint endpoint = { { { 0 } } };
 		memcpy(&endpoint, &in_peer.endpoint, sizeof(in_peer.endpoint));
@@ -89,14 +107,14 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 	}
 
 	if (in_peer.persistent_keepalive_interval != (u16)-1) {
-		const bool send_keepalive = !peer->persistent_keepalive_interval && in_peer.persistent_keepalive_interval && netdev_pub(wg)->flags & IFF_UP;
+		const bool send_keepalive = !peer->persistent_keepalive_interval && in_peer.persistent_keepalive_interval && wg->dev->flags & IFF_UP;
 		peer->persistent_keepalive_interval = (unsigned long)in_peer.persistent_keepalive_interval * HZ;
 		if (send_keepalive)
 			packet_send_keepalive(peer);
 	}
 
-	if (netdev_pub(wg)->flags & IFF_UP)
-		packet_send_queue(peer);
+	if (wg->dev->flags & IFF_UP)
+		packet_send_staged_packets(peer);
 
 	peer_put(peer);
 
@@ -109,8 +127,9 @@ out:
 
 int config_set_device(struct wireguard_device *wg, void __user *user_device)
 {
-	int ret = 0;
+	int ret;
 	size_t i, offset;
+	struct wireguard_peer *peer, *temp;
 	struct wgdevice in_device;
 	void __user *user_peer;
 	bool modified_static_identity = false;
@@ -120,14 +139,18 @@ int config_set_device(struct wireguard_device *wg, void __user *user_device)
 
 	mutex_lock(&wg->device_update_lock);
 
-	if (copy_from_user(&in_device, user_device, sizeof(in_device))) {
-		ret = -EFAULT;
+	ret = -EFAULT;
+	if (copy_from_user(&in_device, user_device, sizeof(in_device)))
+		goto out;
+
+	ret = -EPROTO;
+	if (in_device.version_magic != WG_API_VERSION_MAGIC)
 		goto out;
-	}
 
 	if (in_device.fwmark || (!in_device.fwmark && (in_device.flags & WGDEVICE_REMOVE_FWMARK))) {
 		wg->fwmark = in_device.fwmark;
-		peer_for_each_unlocked(wg, clear_peer_endpoint_src, NULL);
+		peer_for_each (wg, peer, temp, false)
+			socket_clear_peer_endpoint_src(peer);
 	}
 
 	if (in_device.port) {
@@ -143,27 +166,34 @@ int config_set_device(struct wireguard_device *wg, void __user *user_device)
 		noise_set_static_identity_private_key(&wg->static_identity, NULL);
 		modified_static_identity = true;
 	} else if (memcmp(zeros, in_device.private_key, WG_KEY_LEN)) {
+		u8 public_key[NOISE_PUBLIC_KEY_LEN] = { 0 };
+		struct wireguard_peer *peer;
+		/* We remove before setting, to prevent race, which means doing two 25519-genpub ops. */
+		bool unused __attribute((unused)) = curve25519_generate_public(public_key, in_device.private_key);
+		peer = pubkey_hashtable_lookup(&wg->peer_hashtable, public_key);
+		if (peer) {
+			peer_put(peer);
+			peer_remove(peer);
+		}
+
 		noise_set_static_identity_private_key(&wg->static_identity, in_device.private_key);
 		modified_static_identity = true;
 	}
 
-	if (in_device.flags & WGDEVICE_REMOVE_PRESHARED_KEY) {
-		noise_set_static_identity_preshared_key(&wg->static_identity, NULL);
-		modified_static_identity = true;
-	} else if (memcmp(zeros, in_device.preshared_key, WG_KEY_LEN)) {
-		noise_set_static_identity_preshared_key(&wg->static_identity, in_device.preshared_key);
-		modified_static_identity = true;
+	if (modified_static_identity) {
+		peer_for_each (wg, peer, temp, false) {
+			if (!noise_precompute_static_static(peer))
+				peer_remove(peer);
+		}
+		cookie_checker_precompute_device_keys(&wg->cookie_checker);
 	}
 
-	if (modified_static_identity)
-		cookie_checker_precompute_keys(&wg->cookie_checker, NULL);
-
 	for (i = 0, offset = 0, user_peer = user_device + sizeof(struct wgdevice); i < in_device.num_peers; ++i, user_peer += offset) {
 		ret = set_peer(wg, user_peer, &offset);
 		if (ret)
-			break;
+			goto out;
 	}
-
+	ret = 0;
 out:
 	mutex_unlock(&wg->device_update_lock);
 	memzero_explicit(&in_device.private_key, NOISE_PUBLIC_KEY_LEN);
@@ -186,20 +216,6 @@ static inline int use_data(struct data_remaining *data, size_t size)
 	return 0;
 }
 
-static int calculate_ipmasks_size(void *ctx, struct wireguard_peer *peer, union nf_inet_addr ip, u8 cidr, int family)
-{
-	size_t *count = ctx;
-	*count += sizeof(struct wgipmask);
-	return 0;
-}
-
-static size_t calculate_peers_size(struct wireguard_device *wg)
-{
-	size_t len = peer_total_count(wg) * sizeof(struct wgpeer);
-	routing_table_walk_ips(&wg->peer_routing_table, &len, calculate_ipmasks_size);
-	return len;
-}
-
 static int populate_ipmask(void *ctx, union nf_inet_addr ip, u8 cidr, int family)
 {
 	int ret;
@@ -226,10 +242,9 @@ static int populate_ipmask(void *ctx, union nf_inet_addr ip, u8 cidr, int family
 	return ret;
 }
 
-static int populate_peer(struct wireguard_peer *peer, void *ctx)
+static int populate_peer(struct wireguard_peer *peer, struct data_remaining *data)
 {
 	int ret = 0;
-	struct data_remaining *data = ctx;
 	void __user *upeer = data->data;
 	struct wgpeer out_peer;
 	struct data_remaining ipmasks_data = { NULL };
@@ -240,7 +255,11 @@ static int populate_peer(struct wireguard_peer *peer, void *ctx)
 	if (ret)
 		return ret;
 
+	down_read(&peer->handshake.lock);
 	memcpy(out_peer.public_key, peer->handshake.remote_static, NOISE_PUBLIC_KEY_LEN);
+	memcpy(out_peer.preshared_key, peer->handshake.preshared_key, NOISE_SYMMETRIC_KEY_LEN);
+	up_read(&peer->handshake.lock);
+
 	read_lock_bh(&peer->endpoint_lock);
 	if (peer->endpoint.addr.sa_family == AF_INET)
 		out_peer.endpoint.addr4 = peer->endpoint.addr4;
@@ -266,10 +285,10 @@ static int populate_peer(struct wireguard_peer *peer, void *ctx)
 	return ret;
 }
 
-int config_get_device(struct wireguard_device *wg, void __user *udevice)
+int config_get_device(struct wireguard_device *wg, void __user *user_device)
 {
-	int ret = 0;
-	struct net_device *dev = netdev_pub(wg);
+	int ret;
+	struct wireguard_peer *peer, *temp;
 	struct data_remaining peer_data = { NULL };
 	struct wgdevice out_device;
 	struct wgdevice in_device;
@@ -281,38 +300,50 @@ int config_get_device(struct wireguard_device *wg, void __user *udevice)
 
 	mutex_lock(&wg->device_update_lock);
 
-	if (!udevice) {
-		ret = calculate_peers_size(wg);
+	if (!user_device) {
+		ret = peer_total_count(wg) * sizeof(struct wgpeer)
+		    + routing_table_count_nodes(&wg->peer_routing_table) * sizeof(struct wgipmask);
 		goto out;
 	}
 
-	if (copy_from_user(&in_device, udevice, sizeof(in_device))) {
-		ret = -EFAULT;
+	ret = -EFAULT;
+	if (copy_from_user(&in_device, user_device, sizeof(in_device)))
 		goto out;
-	}
 
+	ret = -EPROTO;
+	if (in_device.version_magic != WG_API_VERSION_MAGIC)
+		goto out;
+
+	out_device.version_magic = WG_API_VERSION_MAGIC;
 	out_device.port = wg->incoming_port;
 	out_device.fwmark = wg->fwmark;
-	strncpy(out_device.interface, dev->name, IFNAMSIZ - 1);
-	out_device.interface[IFNAMSIZ - 1] = 0;
+	memcpy(out_device.interface, wg->dev->name, IFNAMSIZ);
 
 	down_read(&wg->static_identity.lock);
 	if (wg->static_identity.has_identity) {
 		memcpy(out_device.private_key, wg->static_identity.static_private, WG_KEY_LEN);
 		memcpy(out_device.public_key, wg->static_identity.static_public, WG_KEY_LEN);
-		memcpy(out_device.preshared_key, wg->static_identity.preshared_key, WG_KEY_LEN);
 	}
 	up_read(&wg->static_identity.lock);
 
 	peer_data.out_len = in_device.peers_size;
-	peer_data.data = udevice + sizeof(struct wgdevice);
-	ret = peer_for_each_unlocked(wg, populate_peer, &peer_data);
+	peer_data.data = user_device + sizeof(struct wgdevice);
+
+	ret = 0;
+	peer_for_each (wg, peer, temp, false) {
+		ret = populate_peer(peer, &peer_data);
+		if (ret)
+			break;
+	}
 	if (ret)
 		goto out;
 	out_device.num_peers = peer_data.count;
 
-	if (copy_to_user(udevice, &out_device, sizeof(out_device)))
-		ret = -EFAULT;
+	ret = -EFAULT;
+	if (copy_to_user(user_device, &out_device, sizeof(out_device)))
+		goto out;
+
+	ret = 0;
 
 out:
 	mutex_unlock(&wg->device_update_lock);
