@@ -3,7 +3,7 @@
 #include "version.h"
 #include "device.h"
 #include "noise.h"
-#include "packets.h"
+#include "queueing.h"
 #include "ratelimiter.h"
 #include "crypto/chacha20poly1305.h"
 #include "crypto/blake2s.h"
@@ -27,35 +27,29 @@ static int __init mod_init(void)
 #endif
 	noise_init();
 
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	ret = packet_init_data_caches();
+	ret = init_crypt_ctx_cache();
 	if (ret < 0)
 		goto err_packet;
-#endif
 
 	ret = device_init();
 	if (ret < 0)
 		goto err_device;
 
-	pr_info("WireGuard " WIREGUARD_VERSION " loaded. See www.wireguard.io for information.\n");
+	pr_info("WireGuard " WIREGUARD_VERSION " loaded. See www.wireguard.com for information.\n");
 	pr_info("Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.\n");
 
 	return 0;
 
 err_device:
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	packet_deinit_data_caches();
+	deinit_crypt_ctx_cache();
 err_packet:
-#endif
 	return ret;
 }
 
 static void __exit mod_exit(void)
 {
 	device_uninit();
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	packet_deinit_data_caches();
-#endif
+	deinit_crypt_ctx_cache();
 	pr_debug("WireGuard unloaded\n");
 }
 
